// -*- Mode: C++ -*-

// ----------------------------------------------------------------------------
//  cab_arc.cpp
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
//  CN[h
// ----------------------------------------------------------------------------

#include "mspack/mspack.h"
#include "cab_arc.h"
#include "mspack_if.h"
#include "map_file.h"
#include "util.h"
#include "seq_num.h"


// ----------------------------------------------------------------------------
//  mscabd_folder bp[NX
// ----------------------------------------------------------------------------

class cab_folder
{
  const struct mscabd_folder  *m_folder;

public:
  cab_folder(void)
       : m_folder(NULL) { }
  cab_folder(const struct mscabd_folder *folder)
       : m_folder(folder) { }
  cab_folder(const cab_folder& src)
       : m_folder(src.m_folder) { }
  ~cab_folder() { }

  int is_end_of_list(void) const { return (NULL == m_folder); }

  const cab_folder next(void) const
    { return cab_folder((NULL != m_folder) ? m_folder->next : NULL); }
  //int compression_type(void) const { return m_folder->comp_type; }
  uint num_blocks(void) const { return m_folder->num_blocks; }

  int compression_method(void) const
    { return MSCABD_COMP_METHOD(m_folder->comp_type); }
  int compression_level(void) const
    { return MSCABD_COMP_LEVEL(m_folder->comp_type); }
};


// ----------------------------------------------------------------------------
//  mscabd_file bp[NX
// ----------------------------------------------------------------------------

class cab_file
{
  friend class cab_decompressor;
  const struct mscabd_file  *m_file;

public:
  cab_file(void)
       : m_file(NULL) { }
  cab_file(const struct mscabd_file *cabd_file)
       : m_file(cabd_file) { }
  cab_file(const cab_file& src)
       : m_file(src.m_file) { }
  ~cab_file() { }

  int is_end_of_list(void) const { return (NULL == m_file); }

  const cab_file next(void) const
    { return cab_file((NULL != m_file) ? m_file->next : NULL); }
  const char *filename(void) const { return m_file->filename; }
  uint length(void) const { return m_file->length; }
  //int attributes(void) const { return m_file->attribs; }
  int hour(void) const { return m_file->time_h; }
  int minute(void) const { return m_file->time_m; }
  int second(void) const { return m_file->time_s; }
  int year(void) const { return m_file->date_y; }
  int month(void) const { return m_file->date_m; }
  int day(void) const { return m_file->date_d; }
  const cab_folder folder(void) const { return cab_folder(m_file->folder); }
  //uint offset(void) const { return m_file->offset; }
};


// ----------------------------------------------------------------------------
//  mscabd_cabinet bp[NX
// ----------------------------------------------------------------------------

class cab_cabinet
{
  friend class cab_decompressor;

  const struct mscabd_cabinet  *m_cabinet;

public:
  cab_cabinet(void)
       : m_cabinet(NULL) { }
  cab_cabinet(const struct mscabd_cabinet *cabinet)
       : m_cabinet(cabinet) { }
  cab_cabinet(const cab_cabinet& src)
       : m_cabinet(src.m_cabinet) { }
  ~cab_cabinet() { }

  int is_end_of_list(void) const { return (NULL == m_cabinet); }

  const cab_cabinet next(void) const
    { return cab_cabinet((NULL != m_cabinet) ? m_cabinet->next : NULL); }

  const char *filename(void) const { return m_cabinet->filename; }
  off_t base_offset(void) const { return m_cabinet->base_offset; }
  uint length(void) const { return m_cabinet->length; }
  const cab_cabinet prev_cabinet(void) const
    { return cab_cabinet(m_cabinet->prevcab); }
  const cab_cabinet next_cabinet(void) const
    { return cab_cabinet(m_cabinet->nextcab); }
  const char *prev_name(void) const { return m_cabinet->prevname; }
  const char *next_name(void) const { return m_cabinet->nextname; }
  const char *prev_info(void) const { return m_cabinet->previnfo; }
  const char *next_info(void) const { return m_cabinet->nextinfo; }
  const cab_file files(void) const { return cab_file(m_cabinet->files); }
  const cab_folder folders(void) const { return cab_folder(m_cabinet->folders); }
  uint set_id(void) const { return m_cabinet->set_id; }
  uint set_index(void) const { return m_cabinet->set_index; }
  //uint header_reserved(void) const { return m_cabinet->header_resv; }
  int flags(void) const { return m_cabinet->flags; }
};


// ----------------------------------------------------------------------------
//  mscab_decompressor bp[NX
// ----------------------------------------------------------------------------

class cab_decompressor
{
  struct mscab_decompressor  *m_decomp;
  cab_cabinet  m_cab;

public:
  cab_decompressor(struct mspack_system *msp_sys = NULL)
       : m_decomp(NULL)
  {
    m_decomp = mspack_create_cab_decompressor(msp_sys);
    set_param(MSCABD_PARAM_DECOMPBUF, 65536);
  }
  ~cab_decompressor()
  {
    close();
    mspack_destroy_cab_decompressor(m_decomp);
  }

  const cab_cabinet open(const char *filename)
  {
    if (NULL != m_decomp)
    {
      close();
      m_cab = cab_cabinet(m_decomp->open(m_decomp, (char *)filename));
    }
    return m_cab;
  }

  void close(void)
  {
    if (NULL != m_decomp && !m_cab.is_end_of_list())
    {
      m_decomp->close(m_decomp, (struct mscabd_cabinet *)(m_cab.m_cabinet));
      m_cab = cab_cabinet(NULL);
    }
  }

  const cab_cabinet search(const char *filename)
  {
    if (NULL != m_decomp)
    {
      close();
      m_cab = cab_cabinet(m_decomp->search(m_decomp, (char *)filename));
    }
    return m_cab;
  }

  int extract(const cab_file& cf, const char *filename)
  {
    if (NULL != m_decomp)
      return m_decomp->extract(m_decomp,
                               (struct mscabd_file *)cf.m_file,
                               (char *)filename);//"axcab_output");
    else
      return 1;
  }

  int set_param(int param, int value)
  {
    if (NULL != m_decomp)
      return m_decomp->set_param(m_decomp, param, value);
    else
      return 1;
  }
};


// ----------------------------------------------------------------------------
//  cabinet A[JCu
// ----------------------------------------------------------------------------

// ---
//  (private) fileInfo  ɒlݒ肷
// ---

void cab_archive::to_fileinfo(fileInfo *fi, const cab_file *file)
{
  // \bh
  {
    bool        level = false;
    const char  *method_str;
    switch (file->folder().compression_method())
    {
    case MSCAB_COMP_NONE: method_str = "none"; break;
    case MSCAB_COMP_MSZIP: method_str = "mszip"; break;
    case MSCAB_COMP_QUANTUM: method_str = "qtm"; level = true; break;
    case MSCAB_COMP_LZX: method_str = "lzx"; level = true; break;
    default: method_str = "unknown"; break;
    }

    if (level)
      wsprintf(fi->method,
               "%s%2d",
               method_str,
               file->folder().compression_level() % 100);
    else
      strncpy(fi->method, method_str, 8);
  }

  fi->position = m_file_num;
  fi->compsize = 1;
  fi->filesize = file->length();
  fi->timestamp = 0;

  fi->path[0] = fi->filename[0] = '\0';
  split_filepath(file->filename(),
                 fi->path, 200,
                 fi->filename, 200);

  fi->crc = 0;
}


// ---
//  A[JCuJ
// ---

int cab_archive::open_archive(const char *arc_name)
{
  close_archive();

  // mspack_system pӂ
  try {
    m_cab_dec = new cab_decompressor(get_mspack_system());
    m_cab_file = new cab_file(NULL);
  } catch (...) {
    return 1;  // memory alloc.
  }

  // A[JCuJ
  cab_cabinet  cab = m_cab_dec->open(arc_name);
  if (cab.is_end_of_list())
  {
    // cabinet TĊJ
    cab = m_cab_dec->search(arc_name);
    if (cab.is_end_of_list())
      return 1;  // m̌`
  }

  *m_cab_file = cab.files();
  m_file_num = 0;
  m_skipped = true;

  return 0;
}


// ---
//  A[JCu
// ---

void cab_archive::close_archive(void)
{
  if (NULL != m_cab_file)
  {
    delete m_cab_file;
    m_cab_file = NULL;
  }
  if (NULL != m_cab_dec)
  {
    m_cab_dec->close();
    delete m_cab_dec;
    m_cab_dec = NULL;
  }
}


// ---
//  i[t@C擾
// ---

int cab_archive::get_next_fileinfo(fileInfo *fi)
{
  if (false == m_skipped)
    *m_cab_file = m_cab_file->next();

  if (m_cab_file->is_end_of_list())
    return -1;

  if (NULL != fi)
    to_fileinfo(fi, m_cab_file);

  m_file_num++;
  m_skipped = false;

  return 0;
}


// ---
//  t@C𒊏o
// ---

int cab_archive::extract(uint8 *buf, size_t buf_size, spi_callback *spicb)
{
  const size_t  BUF_SIZE = MAX_PATH * 2;
  char  fmo_name[BUF_SIZE + 1];

  // t@C}bsOIuWFNgɕt閼O𐶐
#if 0
  strncpy(fmo_name, m_cab_file->filename(), BUF_SIZE);
  fmo_name[BUF_SIZE] = '\0';
  replace_char(fmo_name, '//', '%');
#else
  {
    seq_num_generator  *gen = seq_num_generator::get_instance();
    wsprintf(fmo_name, "%%axcab-%08x%%", gen->generate());
  }
#endif

  file_mapping_object  m_fmo;
  extract_info  ex_info = { buf, buf_size, spicb };
  if (m_fmo.create(fmo_name, sizeof(extract_info)) ||
      m_fmo.write(&ex_info, sizeof(ex_info)))
    return 1;  // memory alloc.

  return m_cab_dec->extract(*m_cab_file, fmo_name);
}


// ---
//  cabinet 肷
// ---

#pragma pack(1)
struct cabinet_header
{
  uint8   m_signature[4];
  uint32  m_resv_1;
  uint32  m_file_size;
  uint32  m_resv_2;
  uint32  m_cffile_offset;
  uint32  m_resv_3;
  uint8   m_minor_ver;
  uint8   m_major_ver;
};
#pragma pack()

int cab_archive::is_cabinet(const uint8 *buf)
{
  const cabinet_header  *hdr = (cabinet_header *)buf;

  if (memcmp("MSCF", hdr->m_signature, 4) ||
      hdr->m_resv_1 || hdr->m_resv_2 || hdr->m_resv_3 ||
      0 == hdr->m_file_size || 0 == hdr->m_cffile_offset ||
      1 < hdr->m_major_ver)
    return 0;
  else
    return 1;
}

int cab_archive::is_cabinet_sfx(HANDLE fh)
{
  if (INVALID_HANDLE_VALUE == fh)
    return 0;

  uint32  sfx_module_size = get_exe_size(fh);
  if (uint32(-1) == sfx_module_size)
    return 0;

  SetFilePointer(fh, sfx_module_size, NULL, FILE_BEGIN);

  DWORD  read_bytes;
  uint8  buf[sizeof(cabinet_header)];
  ReadFile(fh, buf, sizeof(cabinet_header), &read_bytes, NULL);

  return (sizeof(cabinet_header) == read_bytes) ? is_cabinet(buf) : 0;
}
